"""Operations to handle different timezones."""

from __future__ import annotations

from datetime import datetime, tzinfo

try:
    from zoneinfo import ZoneInfo as timezone
except ImportError:
    # Python 3.10+ has zoneinfo built-in, no backports needed
    raise ImportError("zoneinfo not available")


IANA_MAPPING = {
    "IST": "Asia/Kolkata",  # Indian Standard Time (commonly intended)
    "EST": "America/New_York",  # Eastern Standard Time (US)
    "CST": "America/Chicago",  # Central Standard Time (US)
    "PST": "America/Los_Angeles",  # Pacific Standard Time (US)
    "BST": "Europe/London",  # British Summer Time
    "JST": "Asia/Tokyo",  # Japan Standard Time
    "CET": "Europe/Paris",  # Central European Time
    "EET": "Europe/Bucharest",  # Eastern European Time
    # Add more as needed, but always check for ambiguity!
}


def find_timezone(name: str | tzinfo | timezone) -> tzinfo | None:
    """Retrieves the given timezone by name."""
    # check if already converted
    if isinstance(name, (tzinfo, timezone)):
        return name

    # note: IANA tz are not covered by `ZoneInfo` so need to map
    if isinstance(name, str) and name in IANA_MAPPING:
        name = IANA_MAPPING[name]

    # try to convert
    try:
        return timezone(name)
    except Exception:
        return None


def current_timezone() -> tzinfo:
    """Retrieves the currently active timezone."""
    # retrieve the current timezone name
    cur = datetime.now().astimezone().tzname()

    # special cases
    # note: this timezone is not available through the data (summer-time)
    if cur == "CEST":
        cur = "CET"

    return timezone(cur)  # type: ignore[arg-type]



"""All conversion function. This includes conversions between different datetimes."""

from __future__ import annotations

from datetime import date, datetime, time, tzinfo
from logging import Logger
from typing import Any

from dateutil import parser

# try to import various dependend libraries
try:
    from zoneinfo import ZoneInfo as timezone
except ImportError:
    # Python 3.10+ has zoneinfo built-in, no backports needed
    raise ImportError("zoneinfo not available")
try:
    import pandas as pd
    from pandas import DataFrame, Series, to_datetime
    from pandas.api.types import is_datetime64_any_dtype as is_datetime
except Exception:
    pd = None  # type: ignore[assignment]
    Series = None  # type: ignore[assignment,misc]
    DataFrame = None  # type: ignore[assignment,misc]

    def is_datetime(x: Any) -> bool:  # type: ignore[misc] # noqa: ARG001
        """Mock function when pandas is not installed."""
        return False

    def to_datetime(x: Any, **kwargs: Any) -> None:  # type: ignore[no-redef] # noqa: ARG001
        """Mock function when pandas is not installed."""
        return None  # noqa: RET501


try:
    import numpy as np
    from numpy import array as nparray
except Exception:
    np = None  # type: ignore[assignment]
    nparray = None  # type: ignore[assignment]
import contextlib

from pytz import AmbiguousTimeError

from time_helper.const import DATE_FORMATS
from time_helper.timezone import current_timezone, find_timezone


def parse_time(time_str: str, format: str, timezone: tzinfo | timezone | str) -> datetime:
    """Parses the given time based on the format and timezone (if provdied).

    Args:
        time_str: String value that should be parsed
        format: Format to parse the time from
        timezone: Timezone that should be applied (either str or actual timezone)

    Returns:
        (timzone-aware) datetime object
    """
    # check the current timezone
    tz = find_timezone(timezone)

    # update
    dt = time_str if isinstance(time_str, datetime) else datetime.strptime(time_str, format)
    if tz is not None:
        dt = dt.replace(tzinfo=tz)
    return dt


def unix_to_datetime(ts: str | int | float | Any, tz: timezone | str | Any = None) -> datetime:
    """Converts the given objects into a datetime.

    Args:
        ts: `int` or `long` that contains the timestamp
        tz: `pytz.timezone` that is used for localization

    Returns:
        `datetime` object that contains the time
    """
    # check if should be parsed
    if isinstance(ts, str):
        try:
            ts = int(ts)
        except Exception:
            raise ValueError(f"Unable to convert object ({ts}) into a valid int or long item!")
    # check if can be converted
    if isinstance(ts, int):
        # convert to datetime
        dt = datetime.fromtimestamp(ts, tz=timezone("UTC"))
        if tz is not None:
            dt_result = localize_datetime(dt, tz)
            if dt_result is None:
                raise ValueError("Failed to localize datetime")
            dt = dt_result
        else:
            print("WARNING: No timezone given for timestamp, infering 'UTC' as default!")
        return dt
    raise ValueError(f"Given object ({ts}) is not a valid int or long item!")


def any_to_datetime(
    ts: str | datetime | date | Any, logger: Logger | None = None, date_format: str | None = None
) -> datetime | None:
    """Generates a safe datetime from the input information.

    Args:
        ts: object to convert to datetime
        logger: Logging object to output infos
        date_format: Optional string with the date format to use (otherwise will try common ones)

    Returns:
        `datetime` object if converted or `None`
    """
    dt = None

    # check if special case
    if ts is None or isinstance(ts, datetime):
        return ts

    # check if only date
    if isinstance(ts, date) and not isinstance(ts, datetime):
        dt = datetime.combine(ts, datetime.min.time())

    # convert from int or string
    if dt is None:
        with contextlib.suppress(Exception):
            dt = unix_to_datetime(ts)

    # try relevant string formats
    if dt is None and isinstance(ts, str):
        # check for empty string
        if not ts:
            return None

        # FEAT: improve list
        formats = DATE_FORMATS
        if date_format is not None:
            formats = [date_format, *formats]

        try:
            dt = parser.isoparse(ts)
        except Exception:
            # check all formats
            for fmt in formats:
                try:
                    dt = datetime.strptime(ts, fmt)
                    if logger is not None:
                        logger.info(f"Date-Format '{fmt}' worked")
                except Exception:
                    if logger is not None:
                        logger.info(f"Date-Format '{fmt}' did not work")

    # check if only date
    if isinstance(dt, date) and not isinstance(dt, datetime):
        dt = datetime.combine(dt, datetime.min.time())

    # check for additional types
    if pd:
        if isinstance(dt, pd.Timestamp):
            dt = dt.to_pydatetime()
        if dt == pd.NaT:
            return None
    if np:
        if isinstance(dt, np.datetime64):
            dt = dt.astype(datetime)
        if dt == np.nan:
            return None

    if dt is None:
        raise ValueError(f"Unable to parse datetime ({ts})")

    return dt


def convert_to_datetime(
    dt: datetime | date | time, baseline: datetime | None = None, remove_tz: bool = False
) -> datetime:
    """Converts the given data to datetime.

    This might include conversions from date and time data-types

    Args:
        dt: Datetime, time or date object to convert
        baseline: datetime object to retrieve info from or None. Default is datetime.now()
        remove_tz: Defines if timezone information should be removed

    Returns:
        datetime object
    """
    # check baseline
    if baseline is None:
        baseline = datetime.now()

    # check for types
    if isinstance(dt, datetime):
        pass
    elif isinstance(dt, time):
        dt = datetime(baseline.year, baseline.month, baseline.day, dt.hour, dt.minute, dt.second)
    elif isinstance(dt, date):
        dt = datetime(dt.year, dt.month, dt.day, 12, 0)
    else:
        raise ValueError(f"Given datetime data has unkown type ({type(dt)}")

    # check for removal
    if remove_tz:
        dt = dt.replace(tzinfo=None)

    return dt


def localize_datetime(dt: datetime | None, tz: Any | str | timezone | None = None) -> datetime | None:
    """Localizes a datetime to the current timezone.

    Args:
        dt (datetime): Datetime to make aware
        tz (str, timezone): Timezone (either directly or name of the timezone)
    """
    # check if None
    if dt is None:
        return None
    if tz is None:
        tz = current_timezone()

    # update the timezone
    if isinstance(tz, str):
        tz = timezone(tz)

    # check if timezone should be added or converted
    if dt.tzinfo is None:
        return dt.replace(tzinfo=tz)
    return dt.astimezone(tz)


def make_aware(
    dt: datetime | str | Any,
    tz: str | timezone | Any = None,
    force_convert: bool = True,
    col: str | None = None,
) -> datetime | Any:
    """Checks if the current datetime is aware, otherwise make aware.

    Args:
        dt (datetime, str, Any): Datetime to convert
        tz (str, ZoneInfo, Any): Name of the timezone to convert to
        force_convert (bool): Defines if the timezone should be converted if there is already a timezone present
        col (str, None): Column used if the input is pandas Dataframe

    Returns:
        Updated datetime (or pandas object)
    """
    # ensure that data is not none
    if dt is None:
        return None

    # check for pandas
    is_pandas = False
    with contextlib.suppress(NameError, TypeError):
        # Check if pandas types are available and dt is an instance
        is_pandas = isinstance(dt, (Series, DataFrame))

    if is_pandas:
        assert col is not None, "Column is required for pandas objects"
        return make_aware_pandas(dt, col, tz=tz)  # type: ignore[arg-type]

    # make sure dt is datetime
    dt = any_to_datetime(dt)
    if dt is None:
        return None

    # check if already aware
    if dt.tzinfo is not None and (tz is None or force_convert is False):
        return dt

    # check for local timezone (if none provided)
    if tz is None:
        tz = current_timezone()

    # return localized datetime
    return localize_datetime(dt, tz)


def make_aware_pandas(
    df: Series | DataFrame, col: str, format: str | None = None, tz: str | timezone | None = None
) -> Series | DataFrame:
    """This will make the pandas column datetime aware in the specified timezone.

    Defaults the data to the current timezone.

    Args:
        df: DataFrame to convert
        col: name of the column to convert
        format: Default format to try
        timezone: default timezone to convert to

    Returns:
        Updated DataFrame
    """
    # check if pandas is install
    if Series is None or DataFrame is None:
        raise ImportError("Pandas Library is not installed")

    # safty checks
    if df is None:
        return None
    if col is None:
        raise ValueError("Expected column name, but got None")
    if col not in df:
        raise RuntimeError(f"The specified column {col} is not available in the dataframe: {df.columns}")

    # make sure the data is unaware
    if not is_datetime(df[col]):
        # generate format list
        formats = []
        if format is not None:
            formats.append(format)
        formats.extend(DATE_FORMATS)

        # check all formats
        for fmt in formats:
            with contextlib.suppress(Exception):
                df[col] = to_datetime(df[col], format=fmt)

    # TODO: update timezone ensurances
    # ensure timezone
    if not hasattr(df[col].iloc[0], "tzinfo") or not df[col].iloc[0].tzinfo:
        cur_tz_obj = current_timezone()
        cur_tz = getattr(cur_tz_obj, "key", str(cur_tz_obj))
        try:
            df[col] = df[col].dt.tz_localize(cur_tz)
        except AmbiguousTimeError:
            infer_dst = nparray([False] * df.shape[0])
            df[col] = df[col].dt.tz_localize(cur_tz, ambiguous=infer_dst)
    if tz is not None:
        # convert to string (as pandas does not support ZoneInfo)
        if isinstance(tz, timezone):
            tz = tz.key
        df.loc[:, col] = df[col].dt.tz_convert(tz)

    return df


def make_unaware(dt: datetime | Any, tz: str | tzinfo | timezone = "UTC") -> datetime | None:
    """Makes the given timezone unaware in a default timezone.

    Args:
        dt: Datetime to convert
        tz: Timezone to convert to as basetime

    Returns:
        datetime object without timezone info
    """
    # ensure the datetime is safe
    dt = any_to_datetime(dt)

    # check against None values
    if dt is None:
        return None

    # convert the timezone
    result = localize_datetime(dt, tz)
    if result is None:
        return None
    return result.replace(tzinfo=None)



from datetime import datetime, tzinfo

try:
    import zoneinfo
except ImportError:
    import backports.zoneinfo as zoneinfo

from time_helper import current_timezone, find_timezone

LOCAL_TZ = datetime.now().astimezone().tzname()
LOCAL_TZ = "CET" if LOCAL_TZ == "CEST" else LOCAL_TZ


def test_findtz():
    tz = find_timezone("UTC")
    assert type(tz) in (tzinfo, zoneinfo.ZoneInfo)
    assert tz is not None
    assert tz == zoneinfo.ZoneInfo("UTC")

    tz = find_timezone("Asia/Kolkata")
    assert type(tz) in (tzinfo, zoneinfo.ZoneInfo)
    assert tz is not None
    assert tz == zoneinfo.ZoneInfo("Asia/Kolkata")

    tz = find_timezone("foobar")
    assert tz is None

    tz = find_timezone("IST")
    assert tz is not None
    assert tz == zoneinfo.ZoneInfo("Asia/Kolkata")


def test_currenttz():
    tz = current_timezone()
    assert type(tz) in (tzinfo, zoneinfo.ZoneInfo)
    assert type is not None
    assert tz == zoneinfo.ZoneInfo(LOCAL_TZ)



# define the version
__version__ = "0.1.6"

import logging

try:
    from . import const as const
    from .convert import (
        any_to_datetime as any_to_datetime,
    )
    from .convert import (
        convert_to_datetime as convert_to_datetime,
    )
    from .convert import (
        localize_datetime as localize_datetime,
    )
    from .convert import (
        make_aware as make_aware,
    )
    from .convert import (
        make_unaware as make_unaware,
    )
    from .convert import (
        parse_time as parse_time,
    )
    from .convert import (
        unix_to_datetime as unix_to_datetime,
    )
    from .ops import has_timezone as has_timezone
    from .ops import round_time as round_time
    from .ops import time_diff as time_diff
    from .range import create_intervals as create_intervals
    from .range import time_to_interval as time_to_interval
    from .timezone import current_timezone as current_timezone
    from .timezone import find_timezone as find_timezone
    from .wrapper import DateTimeWrapper as DateTimeWrapper
except Exception:
    logging.error("Not all dependencies installed")

__all__ = [
    "DateTimeWrapper",
    "any_to_datetime",
    "const",
    "convert_to_datetime",
    "create_intervals",
    "current_timezone",
    "find_timezone",
    "has_timezone",
    "localize_datetime",
    "make_aware",
    "make_unaware",
    "parse_time",
    "round_time",
    "time_diff",
    "time_to_interval",
    "unix_to_datetime",
]



"""This contains all range operations (such as intervals)."""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any

from time_helper import any_to_datetime, convert_to_datetime


def time_to_interval(
    dt: Any,
    offset: int | tuple[int, int] | list[int] = 12,
    baseline: datetime | None = None,
    zero_center: bool = True,
    normalize: bool = True,
) -> float:
    """Converts a datetime value into an interval along the day.

    In case of normalization the data is ranged from 0 to 1 (for the timestamp +- offset)
    If zero_center is enabled this range shifts to [-.5, .5]

    Args:
        dt: The datetime to convert
        offset: Number of hours to add to both ends to avoid errors due to overcasts.
            This can also be a tuple to have asymmetric offset
        baseline: Datetime that is used as baseline (if None take day from the dt)
        zero_center: Defines if the middle of the time range should be 0 centered
            (meaning everything before is negative values)

    Returns:
        Float value of the time position - if normalized a value between 0 and 1 (1 = last possible time) - otherwise a value in minutes
    """
    # convert to unaware
    dt_base = convert_to_datetime(baseline, None, True) if baseline else None
    dt_uw = convert_to_datetime(dt, baseline, True)
    dt_base = baseline if baseline else dt_uw

    # retrieve offset
    if isinstance(offset, tuple) or isinstance(offset, list):
        offset_start = offset[0]
        offset_end = offset[1]
    else:
        offset_start = offset
        offset_end = offset

    # compute the total time
    total_min = (24 + offset_start + offset_end) * 60
    total_end = datetime(dt_base.year, dt_base.month, dt_base.day, 23, 59) + timedelta(hours=offset_end, minutes=1)

    # measure time by distance to the total end
    dt_min = total_min - ((total_end - dt_uw).total_seconds() / 60)

    # check for centering
    if zero_center:
        dt_min -= total_min / 2

    # check for normalization
    if normalize:
        dt_min /= total_min

    return dt_min


def create_intervals(
    start: Any,
    end: Any = None,
    interval: int | float | timedelta = 6,
    round_days: bool = False,
    skip: timedelta = timedelta(seconds=1),
) -> list[tuple[datetime, datetime]]:
    """Generates an array of interval tuples for the given date range.

    Args:
        start (datetime): The to start at
        end (datetime): The time to end at
        interval_days (int): Number of days for each interval (or timedelta)
        round_days (bool): If the time from the input should be preserved or rounded to whole days

    Returns:
        List of datetime tuples (note that these are timezone aware) of start and end date
    """
    # update the start and end dates
    start_date = any_to_datetime(start)
    if start_date is None:
        raise ValueError("Failed to parse start date")

    if end is None:
        end_date = datetime.utcnow()
    else:
        parsed_end = any_to_datetime(end)
        if parsed_end is None:
            raise ValueError("Failed to parse end date")
        end_date = parsed_end

    if round_days:
        start_date = start_date.replace(hour=0, minute=0, second=0, microsecond=0)
        end_date = end_date.replace(hour=23, minute=59, second=59, microsecond=999999)

    # check interval
    if not isinstance(interval, timedelta):
        if isinstance(interval, int) or isinstance(interval, float):
            interval = timedelta(days=interval)
        else:
            raise ValueError("Invalid interval passed")

    # convert to range
    date_range: list[tuple[datetime, datetime]] = []
    d_start: datetime = start_date
    while d_start < end_date:
        # update new item
        d_end = min(end_date, d_start + interval)
        if d_end - d_start > skip:
            date_range.append((d_start, d_end))

        # update the time
        d_start = d_start + interval

    return date_range



"""Contains a range of operations on different datetimes."""

from __future__ import annotations

from datetime import date, datetime, timedelta
from typing import Any

try:
    from pandas import DataFrame, Series
    from pandas.api.types import is_datetime64_any_dtype as is_datetime
except Exception:
    Series = None  # type: ignore[assignment,misc]
    DataFrame = None  # type: ignore[assignment,misc]

    def is_datetime(x: Any) -> bool:  # type: ignore[misc] # noqa: ARG001
        """Mock function when pandas is not installed."""
        return False


from time_helper.convert import any_to_datetime, localize_datetime, make_aware


def has_timezone(df: Series | DataFrame, col: str | None = None) -> bool:
    """Checks if a given pandas object has a timezone.

    Args:
        df: Pandas DataFrame or Series to check
        col: Column name if df is a DataFrame

    Returns:
        True if the object has timezone information, False otherwise
    """
    # perform security
    if df is None:
        raise ValueError("Expected a dataframe but got None")
    if isinstance(df, DataFrame):
        if col is None:
            raise ValueError("Expected a column name, but got None")
        if col not in df.columns:
            raise ValueError(f"The provided column {col} is not in the dataframe {df.columns}")
        if not is_datetime(df[col]):
            raise ValueError("Specified column is not a datetime object!")
    elif not is_datetime(df):
        raise ValueError("Provided series is not a datetime object!")
    if len(df) == 0:
        raise ValueError("The Dataframe is empty")

    # perform checks
    df_col = df if isinstance(df, Series) else df[col]
    obj = df_col.iloc[0]
    if not hasattr(obj, "tzinfo"):
        return False
    return obj.tzinfo is not None


def time_diff(dt1: datetime, dt2: datetime, tz: str | Any = None) -> timedelta:
    """Allows to compute the difference between datetimes of different timezones.

    Calculates dt1 - dt2

    Args:
        dt1 (datetime): First datetime in the diff
        dt2 (datetime): Second datetime in the diff
        tz (str): Name of the default timezone to use

    Returns:
        timedelta
    """
    # check for unaware
    if dt1.tzinfo is None and dt2.tzinfo is None:
        return dt1 - dt2

    # update timezones
    if dt1.tzinfo is None:
        dt1_aware = make_aware(dt1, tz, force_convert=False)
        if not isinstance(dt1_aware, datetime):
            raise ValueError("Failed to make dt1 aware")
        dt1 = dt1_aware
    if dt2.tzinfo is None:
        dt2_aware = make_aware(dt2, tz, force_convert=False)
        if not isinstance(dt2_aware, datetime):
            raise ValueError("Failed to make dt2 aware")
        dt2 = dt2_aware

    # bring both datetimes to same timezone
    dt1_utc = localize_datetime(dt1, "UTC")
    dt2_utc = localize_datetime(dt2, "UTC")
    if dt1_utc is None or dt2_utc is None:
        raise ValueError("Failed to localize to UTC")
    dt1 = dt1_utc
    dt2 = dt2_utc

    # calculate
    return dt1 - dt2


def round_time(dt: datetime, freq: str = "D", max_out: bool = False) -> datetime | None:
    """Rounds the given timestamp to the start or end of the day.

    Args:
        timestamp (datetime): Datetime that should be rounded
        freq (str): Frequency to round it to (options are S, M, H, D, W, m, Y)
        max_out (bool): Defines if remaining attributes should be zeroed (False) or maxed out (True)

    Returns:
        Updated datetime
    """
    # check if value is provided
    if not dt:
        return None

    # ensure value is a datetime
    if not isinstance(dt, datetime):
        try:
            dt = any_to_datetime(dt)
        except Exception:
            raise ValueError(f"Could not parse Timestamp: {dt}")

    # check for date
    if not isinstance(dt, datetime) and isinstance(dt, date) and freq in ["H", "M", "S"]:
        raise ValueError("Got a date, but frequency requires datetime")

    # check for week special case
    if freq == "W":
        day = dt.weekday()
        dt = dt + timedelta(days=6 - day) if max_out else dt - timedelta(days=day)

    # update time to get most out of day
    items = {"microsecond": 999999 if max_out else 0}
    if freq not in ["S"]:
        items["second"] = 59 if max_out else 0
    if freq not in ["S", "M"]:
        items["minute"] = 59 if max_out else 0
    if freq not in ["S", "M", "H"]:
        items["hour"] = 23 if max_out else 0
    if freq in ["Y", "m"]:
        day_in_month = (dt.replace(month=dt.month % 12 + 1, day=1) - timedelta(days=1)).day
        items["day"] = day_in_month if max_out else 1
    if freq in ["Y"]:
        items["month"] = 12 if max_out else 1
        items["day"] = 31 if max_out else items["day"]
    dt = dt.replace(**items)  # type: ignore[arg-type]

    return dt



from datetime import datetime, timedelta

import pandas as pd
import pytest

from time_helper import has_timezone, localize_datetime, round_time, time_diff


def test_diff():
    # setup the data
    date1 = datetime(2021, 1, 1, 15, 10, 30)
    date2 = datetime(2021, 1, 10, 18, 20, 50)
    orig_diff = timedelta(days=9, hours=3, minutes=10, seconds=20)

    # vanilla test
    diff1 = time_diff(date2, date1)
    assert diff1 == orig_diff
    diff2 = time_diff(date1, date2)
    assert diff2 == -orig_diff

    # localize to first timezone
    date1_utc = localize_datetime(date1, "UTC")
    diff1 = time_diff(date2, date1_utc, "UTC")
    assert diff1 == orig_diff
    diff2 = time_diff(date1_utc, date2, "UTC")
    assert diff2 == -orig_diff

    # localize further should not change inherit time
    date1_cal = localize_datetime(date1_utc, "Asia/Calcutta")
    diff1 = time_diff(date2, date1_utc, "UTC")
    assert diff1 == orig_diff

    # test different timezones (times are then compared in utc, so diff should remove 1 hour)
    date2_cet = localize_datetime(date2, "Europe/Berlin")
    diff1 = time_diff(date2_cet, date1_utc)
    assert diff1 == orig_diff - timedelta(hours=1)

    # convert the other timezone
    date1_cal = localize_datetime(date1, "Asia/Calcutta")
    diff1 = time_diff(date2_cet, date1_cal)
    assert diff1 == orig_diff + timedelta(hours=4, minutes=30)


def test_pandas_timezone():
    df = pd.DataFrame(
        [
            [
                pd.Timestamp("2020-06-06").tz_localize("Europe/Berlin"),
                "foo",
            ],
            [pd.Timestamp("2020-06-06").tz_localize("Europe/Berlin"), "bar"],
        ],
        columns=["date", "text"],
    )

    # check error cases
    with pytest.raises(ValueError):
        has_timezone(None, None)
    with pytest.raises(ValueError):
        has_timezone(df, None)
    with pytest.raises(ValueError):
        has_timezone(df, "no_col")
    with pytest.raises(ValueError):
        has_timezone(df, "text")

    # check datetime cases
    val = has_timezone(df, "date")
    assert val is True

    # update dataframej
    df = pd.DataFrame(
        [
            [
                pd.Timestamp("2020-06-06"),
                "foo",
            ],
            [pd.Timestamp("2020-06-06"), "bar"],
        ],
        columns=["date", "text"],
    )
    val = has_timezone(df, "date")
    assert val is False


def test_round_time():
    dt = datetime(2022, 2, 10, 13, 30, 54)

    dt_out = round_time(dt, "M", max_out=False)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-02-10T13:30:00"
    dt_out = round_time(dt, "M", max_out=True)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-02-10T13:30:59.999999"

    dt_out = round_time(dt, "H", max_out=False)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-02-10T13:00:00"

    dt_out = round_time(dt, "D", max_out=False)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-02-10T00:00:00"

    dt_out = round_time(dt, "W", max_out=False)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-02-07T00:00:00"
    dt_out = round_time(dt, "W", max_out=True)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-02-13T23:59:59.999999"

    dt_out = round_time(dt, "m", max_out=False)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-02-01T00:00:00"
    dt_out = round_time(dt, "m", max_out=True)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-02-28T23:59:59.999999"

    dt_out = round_time(dt, "Y", max_out=False)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-01-01T00:00:00"
    dt_out = round_time(dt, "Y", max_out=True)
    assert type(dt_out) == datetime
    assert dt_out.isoformat() == "2022-12-31T23:59:59.999999"



"""Wrapper Class for different time objects (including dates and datetimes).

This makes many of the functions first class citizens and can easily expose the datetime
"""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any

from time_helper.convert import any_to_datetime
from time_helper.ops import time_diff


class DateTimeWrapper:
    """Wrapper class for datetime objects with additional operations."""

    def __init__(self, dt: Any) -> None:
        self.dt = any_to_datetime(dt)

    def __call__(self, *args: Any, **kwds: Any) -> datetime | None:
        """Return the wrapped datetime object."""
        if not args and not kwds:
            return self.dt
        # TODO: call action on the internal datetime object
        return self.dt

    # TODO: overload operators
    def __sub__(self, other: Any) -> timedelta:
        # make sure to convert
        if not isinstance(other, DateTimeWrapper):
            other = DateTimeWrapper(other)

        # ensure to compute
        if self.dt is None or other.dt is None:
            raise ValueError("Cannot compute time diff with None datetime")
        return time_diff(self.dt, other.dt)



"""Test Various Convert Functions"""

from datetime import datetime, timedelta

try:
    import zoneinfo
except ImportError:
    import backports.zoneinfo as zoneinfo

import pandas as pd
import pytest

from time_helper import has_timezone, localize_datetime, make_aware, make_unaware, parse_time, unix_to_datetime

LOCAL_TZ = datetime.now().astimezone().tzname()
LOCAL_TZ = "CET" if LOCAL_TZ == "CEST" else LOCAL_TZ


def test_parse():
    date_str = "2021-09-15"
    format = "%Y-%m-%d"
    orig_date = datetime(2021, 9, 15, tzinfo=zoneinfo.ZoneInfo("UTC"))

    # test the parsing
    dt = parse_time(date_str, format, "UTC")
    assert dt is not None
    assert dt == orig_date
    assert dt.tzinfo.tzname(None) == zoneinfo.ZoneInfo("UTC").tzname(None)

    date_str = "2021-09-15_20:14:50"
    format = "%Y-%m-%d_%H:%M:%S"
    orig_date = datetime(2021, 9, 15, 20, 14, 50, tzinfo=zoneinfo.ZoneInfo(LOCAL_TZ))

    # test the parsing
    dt = parse_time(date_str, format, LOCAL_TZ)
    assert dt is not None
    assert dt == orig_date
    assert dt.tzinfo.tzname(None) == zoneinfo.ZoneInfo(LOCAL_TZ).tzname(None)

    # test error case
    with pytest.raises(ValueError):
        parse_time("1020-13-32_12:34:21", format, "UTC")


def test_any():
    # TODO: implement
    pass


def test_unix():
    # pair to check
    unix = 1634394762
    date = datetime(2021, 10, 16, 14, 32, 42, tzinfo=zoneinfo.ZoneInfo("GMT"))

    # run conversion
    conv_date = unix_to_datetime(unix, "GMT")
    assert conv_date is not None
    assert conv_date.tzinfo is not None
    assert conv_date == date

    # test with string
    date_loc = localize_datetime(date, LOCAL_TZ)
    conv_date = unix_to_datetime(str(unix), LOCAL_TZ)
    assert conv_date is not None
    assert conv_date.tzinfo is not None
    assert conv_date == date_loc

    # test error case
    with pytest.raises(ValueError):
        unix_to_datetime("FOO", "UTC")


def test_localize():
    date = datetime(2020, 10, 15, 20, 15, 13)
    assert date.tzinfo is None

    # makes the date aware
    loc_date = localize_datetime(date)
    assert loc_date.tzinfo.tzname(None) == zoneinfo.ZoneInfo(LOCAL_TZ).tzname(None)
    assert date.date() == loc_date.date()
    assert date.time() == loc_date.time()

    # ensures that the aware date can be converted in timezone
    loc_date_2 = localize_datetime(loc_date, "Asia/Kolkata")
    assert loc_date_2.tzinfo.tzname(None) == zoneinfo.ZoneInfo("Asia/Kolkata").tzname(None)
    assert abs(date.date() - loc_date_2.date()) <= timedelta(days=1)
    assert date.time() != loc_date_2.time()

    # makes sure that conversion back restores original timezone
    loc_date_3 = localize_datetime(loc_date_2, LOCAL_TZ)
    assert loc_date_3.tzinfo.tzname(None) == zoneinfo.ZoneInfo(LOCAL_TZ).tzname(None)
    assert date.date() == loc_date_3.date()
    assert date.time() == loc_date_3.time()

    # make sure that timestamp gets changed
    loc_date_4 = localize_datetime(datetime(2021, 10, 21, 5, 44, 18), "UTC")
    loc_date_5 = localize_datetime(datetime(2021, 10, 21, 5, 44, 18), "Europe/Berlin")
    assert loc_date_4.timestamp() > loc_date_5.timestamp()
    assert loc_date_4.timestamp() == loc_date_5.timestamp() + (60 * 60 * 2)


def test_aware():
    # create basic date
    date = datetime(2020, 10, 15, 20, 15, 13)
    assert date.tzinfo is None

    # ensure that datetime gets added
    loc_date = make_aware(date, "Europe/Berlin")
    assert loc_date is not None
    assert loc_date.tzinfo is not None
    assert date.date() == loc_date.date()
    assert date.time() == loc_date.time()
    assert loc_date.tzinfo.tzname(None) == zoneinfo.ZoneInfo("Europe/Berlin").tzname(None)

    # check to make aware timezone aware
    loc_date = make_aware(loc_date, "Europe/Berlin")
    assert loc_date is not None
    assert loc_date.tzinfo is not None
    assert date.date() == loc_date.date()
    assert date.time() == loc_date.time()
    assert loc_date.tzinfo.tzname(None) == zoneinfo.ZoneInfo("Europe/Berlin").tzname(None)

    # check switch of timezone
    pre_date = localize_datetime(loc_date, "Asia/Calcutta")
    loc_date2 = make_aware(loc_date, "Asia/Calcutta")
    assert pre_date == loc_date2
    assert loc_date2 is not None
    assert loc_date2.tzinfo is not None
    assert date.date() == loc_date2.date()
    assert (date + timedelta(hours=3, minutes=30)).time() == loc_date2.time()
    assert loc_date2.tzinfo.tzname(None) == zoneinfo.ZoneInfo("Asia/Calcutta").tzname(None)

    # make none check
    loc_date = make_aware(None)
    assert loc_date is None

    # test pandas functions
    df = pd.DataFrame(
        [
            [
                pd.Timestamp("2020-06-06"),
                "foo",
            ],
            [pd.Timestamp("2020-06-06"), "bar"],
        ],
        columns=["date", "text"],
    )

    # check error cases
    val = make_aware(None, col=None)
    assert val is None
    with pytest.raises(AssertionError):
        make_aware(df, col=None)
    with pytest.raises(RuntimeError):
        make_aware(df, col="no_col")

    df_new = make_aware(df, col="date")
    assert has_timezone(df_new, "date") is True

    df = pd.DataFrame(
        [
            [
                "2020-06-06",
                "foo",
            ],
            ["2020-06-06", "bar"],
        ],
        columns=["date", "text"],
    )
    df_new = make_aware(df, col="date")
    assert has_timezone(df_new, "date") is True

    df_none = make_aware("")
    assert df_none is None

    # TODO: add additional test cases here


def test_unaware():
    # create data
    date1 = datetime(2021, 1, 1, 15, 10, 30)
    date1_utc = localize_datetime(date1, "UTC")
    date1_cet = localize_datetime(date1, "CET")

    assert date1_cet.timestamp() != date1_utc.timestamp()

    dt = make_unaware(date1)
    assert dt == date1

    dt = make_unaware(date1_utc)
    assert dt == date1

    dt = make_unaware(date1_cet)
    assert dt == date1 - timedelta(hours=1)

    dt = make_unaware(date1_cet, "UTC")
    assert dt == date1 - timedelta(hours=1)

    dt = make_unaware("2021-07-12")
    assert dt == datetime(2021, 7, 12)

    dt = make_unaware("")
    assert dt is None



from datetime import datetime, timedelta

from time_helper import time_to_interval


def test_time_to_interval():
    """Tests if the conversion is correct"""
    dt = datetime(2020, 9, 23, 12, 00)
    iv = time_to_interval(dt, 0)
    assert iv == 0

    iv = time_to_interval(dt, 0, zero_center=False, normalize=True)
    assert iv == 0.5

    iv = time_to_interval(dt, 12)
    assert iv == 0

    iv = time_to_interval(dt, 12, zero_center=False, normalize=True)
    assert iv == 0.5

    iv = time_to_interval(dt, 12, zero_center=False, normalize=False)
    assert iv == 24 * 60

    # test time after the day
    dt = datetime(2020, 9, 24, 6, 00)
    base = dt - timedelta(hours=12)
    assert base.day == 23

    iv = time_to_interval(dt, 12, baseline=base, zero_center=False, normalize=True)
    assert iv == 42 / 48

    iv = time_to_interval(dt, 12, baseline=base, zero_center=False, normalize=False)
    assert iv == 42 * 60

    iv = time_to_interval(dt, 12, baseline=base, zero_center=True, normalize=True)
    assert iv == 18 / 48

    # test time before the day
    dt = datetime(2020, 9, 22, 22, 00)
    base = dt + timedelta(hours=12)
    assert base.day == 23

    iv = time_to_interval(dt, 12, baseline=base, zero_center=False, normalize=True)
    assert iv == 10 / 48

    iv = time_to_interval(dt, 12, baseline=base, zero_center=False, normalize=False)
    assert iv == 10 * 60

    iv = time_to_interval(dt, 12, baseline=base, zero_center=True, normalize=True)
    assert iv == -14 / 48

    # test async offset
    dt = datetime(2020, 9, 24, 6, 00)
    base = dt - timedelta(hours=12)
    assert base.day == 23

    iv = time_to_interval(dt, (6, 12), baseline=base, zero_center=False, normalize=True)
    assert iv == 36 / 42

    iv = time_to_interval(dt, (12, 6), baseline=base, zero_center=False, normalize=False)
    assert iv == 42 * 60

    iv = time_to_interval(dt, (6, 12), baseline=base, zero_center=False, normalize=False)
    assert iv == 36 * 60

    iv = time_to_interval(dt, (6, 12), baseline=base, zero_center=True, normalize=True)
    assert iv == 15 / 42


def test_create_interval():
    # TODO: implement
    pass
